@@ -149,7 +149,7 @@ class PdfFileWriter(object):
                 value = self._sweepIndirectReferences(externMap, value)
                 if value == None:
                     print objects, value, origvalue
-                if hasattr(value, "has_key") and value.has_key("__streamdata__"):
+                if isinstance(value, StreamObject):
                     # a dictionary value is a stream.  streams must be indirect
                     # objects, so we need to change this value.
                     value = self._addObject(value)
@@ -158,7 +158,7 @@ class PdfFileWriter(object):
         elif isinstance(data, ArrayObject):
             for i in range(len(data)):
                 value = self._sweepIndirectReferences(externMap, data[i])
-                if hasattr(value, "has_key") and value.has_key("__streamdata__"):
+                if isinstance(value, StreamObject):
                     # an array value is a stream.  streams must be indirect
                     # objects, so we need to change this value
                     value = self._addObject(value)
@@ -209,6 +209,23 @@ class PdfFileReader(object):
         self.read(stream)
         self.stream = stream
 
+    def getDocumentInfo(self):
+        """
+        Retrieves the PDF file's document information dictionary, if it
+        exists.  Returns a DocumentInformation instance, or None.
+        Note that some PDF files use metadata streams instead of docinfo
+        dictionaries, and these metadata streams will not be accessed by this
+        function.
+
+        Stability: Added in v1.6, will exist for all v1.x releases.
+        """
+        if not self.trailer.has_key("/Info"):
+            return None
+        obj = self.getObject(self.trailer['/Info'])
+        retval = DocumentInformation()
+        retval.update(obj)
+        return retval
+
     def getNumPages(self):
         """
         Returns the number of pages in this PDF file.
@@ -274,7 +291,7 @@ class PdfFileReader(object):
             objStm = self.getObject(IndirectObject(stmnum, 0, self))
             assert objStm['/Type'] == '/ObjStm'
             assert idx < objStm['/N']
-            streamData = StringIO(filters.decodeStreamData(objStm))
+            streamData = StringIO(objStm.getData())
             for i in range(objStm['/N']):
                 objnum = NumberObject.readFromStream(streamData)
                 readNonWhitespace(streamData)
@@ -312,7 +329,7 @@ class PdfFileReader(object):
 
     def read(self, stream):
         # start at the end:
-        stream.seek(-2, 2)
+        stream.seek(-1, 2)
         line = ''
         while not line:
             line = self.readNextEndLine(stream)
@@ -338,25 +355,38 @@ class PdfFileReader(object):
                 assert ref[:3] == "ref"
                 readNonWhitespace(stream)
                 stream.seek(-1, 1)
-                num = readObject(stream, self)
-                readNonWhitespace(stream)
-                stream.seek(-1, 1)
-                size = readObject(stream, self)
-                readNonWhitespace(stream)
-                stream.seek(-1, 1)
-                cnt = 0
-                while cnt < size:
-                    line = stream.readline()
-                    offset, generation = line[:16].split(" ")
-                    offset, generation = int(offset), int(generation)
-                    if not self.xref.has_key(generation):
-                        self.xref[generation] = {}
-                    self.xref[generation][num] = offset
-                    cnt += 1
-                    num += 1
-                readNonWhitespace(stream)
-                stream.seek(-1, 1)
-                assert stream.read(7) == "trailer"
+                while 1:
+                    num = readObject(stream, self)
+                    readNonWhitespace(stream)
+                    stream.seek(-1, 1)
+                    size = readObject(stream, self)
+                    readNonWhitespace(stream)
+                    stream.seek(-1, 1)
+                    cnt = 0
+                    while cnt < size:
+                        line = stream.read(20)
+                        offset, generation = line[:16].split(" ")
+                        offset, generation = int(offset), int(generation)
+                        if not self.xref.has_key(generation):
+                            self.xref[generation] = {}
+                        if self.xref[generation].has_key(num):
+                            # It really seems like we should allow the last
+                            # xref table in the file to override previous
+                            # ones. Since we read the file backwards, assume
+                            # any existing key is already set correctly.
+                            pass
+                        else:
+                            self.xref[generation][num] = offset
+                        cnt += 1
+                        num += 1
+                    readNonWhitespace(stream)
+                    stream.seek(-1, 1)
+                    trailertag = stream.read(7)
+                    if trailertag != "trailer":
+                        # more xrefs!
+                        stream.seek(-7, 1)
+                    else:
+                        break
                 readNonWhitespace(stream)
                 stream.seek(-1, 1)
                 newTrailer = readObject(stream, self)
@@ -367,14 +397,14 @@ class PdfFileReader(object):
                     startxref = newTrailer[NameObject("/Prev")]
                 else:
                     break
-            else:
+            elif x.isdigit():
                 # PDF 1.5+ Cross-Reference Stream
                 stream.seek(-1, 1)
                 idnum, generation = self.readObjectHeader(stream)
                 xrefstream = readObject(stream, self)
                 assert xrefstream["/Type"] == "/XRef"
                 self.cacheIndirectObject(generation, idnum, xrefstream)
-                streamData = StringIO(filters.decodeStreamData(xrefstream))
+                streamData = StringIO(xrefstream.getData())
                 num, size = xrefstream.get("/Index", [0, xrefstream.get("/Size")])
                 entrySizes = xrefstream.get("/W")
                 cnt = 0
@@ -416,6 +446,20 @@ class PdfFileReader(object):
                     startxref = xrefstream["/Prev"]
                 else:
                     break
+            else:
+                # bad xref character at startxref.  Let's see if we can find
+                # the xref table nearby, as we've observed this error with an
+                # off-by-one before.
+                stream.seek(-11, 1)
+                tmp = stream.read(20)
+                xref_loc = tmp.find("xref")
+                if xref_loc != -1:
+                    startxref -= (10 - xref_loc)
+                    continue
+                else:
+                    # no xref table found at specified location
+                    assert False
+                    break
 
     def readNextEndLine(self, stream):
         line = ""
@@ -468,6 +512,7 @@ def addRectangleAccessor(klass, propname, name, fallback, docs):
 
 class PageObject(DictionaryObject):
     def __init__(self, pdf):
+        DictionaryObject.__init__(self)
         self.pdf = pdf
 
     def rotateClockwise(self, angle):
@@ -510,28 +555,28 @@ class PageObject(DictionaryObject):
         return newRes, renameRes
     _mergeResources = staticmethod(_mergeResources)
 
-    def _contentStreamRename(stream, rename):
+    def _contentStreamRename(stream, rename, pdf):
         if not rename:
-            if isinstance(stream, ArrayObject):
-                return stream
-            else:
-                return [stream]
-        stream = ContentStream(stream)
-        newdata = StringIO()
+            return stream
+        stream = ContentStream(stream, pdf)
         for operands,operator in stream.operations:
-            for op in operands:
+            for i in range(len(operands)):
+                op = operands[i]
                 if isinstance(op, NameObject):
-                    op = rename.get(op, op)
-                op.writeToStream(newdata)
-                newdata.write(" ")
-            newdata.write(operator)
-            newdata.write("\n")
-        retval = DictionaryObject()
-        retval["__streamdata__"] = newdata.getvalue()
-        retval[NameObject('/Length')] = NumberObject(len(retval["__streamdata__"]))
-        return [retval]
+                    operands[i] = rename.get(op, op)
+        return stream
     _contentStreamRename = staticmethod(_contentStreamRename)
 
+    def _pushPopGS(contents, pdf):
+        # adds a graphics state "push" and "pop" to the beginning and end
+        # of a content stream.  This isolates it from changes such as 
+        # transformation matricies.
+        stream = ContentStream(contents, pdf)
+        stream.operations.insert(0, [[], "q"])
+        stream.operations.append([[], "Q"])
+        return stream
+    _pushPopGS = staticmethod(_pushPopGS)
+
     def mergePage(self, page2):
         """
         Merges the content streams of two pages into one.  Resource
@@ -566,17 +611,31 @@ class PageObject(DictionaryObject):
         newContentArray = ArrayObject()
 
         originalContent = self["/Contents"].getObject()
-        if isinstance(originalContent, ArrayObject):
-            newContentArray.extend(originalContent)
-        else:
-            newContentArray.append(originalContent)
+        newContentArray.append(PageObject._pushPopGS(originalContent, self.pdf))
 
         page2Content = page2['/Contents'].getObject()
-        newContentArray.extend(PageObject._contentStreamRename(page2Content, rename))
+        page2Content = PageObject._contentStreamRename(page2Content, rename, self.pdf)
+        page2Content = PageObject._pushPopGS(page2Content, self.pdf)
+        newContentArray.append(page2Content)
 
-        self[NameObject('/Contents')] = newContentArray
+        self[NameObject('/Contents')] = ContentStream(newContentArray, self.pdf)
         self[NameObject('/Resources')] = newResources
 
+    def compressContentStreams(self):
+        """
+        Join all content streams and apply a FlateDecode filter to decrease
+        the stream's size.
+
+        Stability: Added in v1.6, will exist for all v1.x releases thereafter.
+        However, if content stream compression is ever handled in a different
+        and/or more transparent way, this function may not do anything.
+        """
+        content = self["/Contents"].getObject()
+        if not isinstance(content, ContentStream):
+            content = ContentStream(content, self.pdf)
+        self[NameObject("/Contents")] = content.flateEncode()
+
+
 addRectangleAccessor(PageObject, "mediaBox", "/MediaBox", (),
         """A rectangle, expressed in default user space units, defining the
         boundaries of the physical medium on which the page is intended to be
@@ -615,34 +674,141 @@ addRectangleAccessor(PageObject, "artBox", "/ArtBox", ("/CropBox",
         thereafter.""")
 
 
-class ContentStream(DictionaryObject):
-    def __init__(self, stream):
+class ContentStream(DecodedStreamObject):
+    def __init__(self, stream, pdf):
+        self.pdf = pdf
         self.operations = []
-        self.__parseContentStream(stream)
-
-    def __parseContentStream(self, stream):
-        # stream may be an array of streams to concatenate.
+        # stream may be a StreamObject or an ArrayObject containing
+        # multiple StreamObjects to be cat'd together.
         stream = stream.getObject()
         if isinstance(stream, ArrayObject):
             data = ""
             for s in stream:
-                data += filters.decodeStreamData(s.getObject())
+                data += s.getObject().getData()
             stream = StringIO(data)
         else:
-            stream = StringIO(filters.decodeStreamData(stream))
+            stream = StringIO(stream.getData())
+        self.__parseContentStream(stream)
+
+    def __parseContentStream(self, stream):
+        # file("f:\\tmp.txt", "w").write(stream.read())
+        stream.seek(0, 0)
         operands = []
         while True:
             peek = readNonWhitespace(stream)
             if peek == '':
                 break
             stream.seek(-1, 1)
-            if peek.isalpha():
-                operator = readUntilWhitespace(stream)
-                self.operations.append((operands, operator))
-                operands = []
+            if peek.isalpha() or peek == "'" or peek == "\"":
+                operator = readUntilWhitespace(stream, maxchars=2)
+                if operator == "BI":
+                    # begin inline image - a completely different parsing
+                    # mechanism is required, of course... thanks buddy...
+                    assert operands == []
+                    ii = self._readInlineImage(stream)
+                    self.operations.append((ii, "INLINE IMAGE"))
+                else:
+                    self.operations.append((operands, operator))
+                    operands = []
             else:
                 operands.append(readObject(stream, None))
 
+    def _readInlineImage(self, stream):
+        # begin reading just after the "BI" - begin image
+        # first read the dictionary of settings.
+        settings = DictionaryObject()
+        while True:
+            tok = readNonWhitespace(stream)
+            stream.seek(-1, 1)
+            if tok == "I":
+                # "ID" - begin of image data
+                break
+            key = readObject(stream, self.pdf)
+            tok = readNonWhitespace(stream)
+            stream.seek(-1, 1)
+            value = readObject(stream, self.pdf)
+            settings[key] = value
+        # left at beginning of ID
+        tmp = stream.read(3)
+        assert tmp[:2] == "ID"
+        data = ""
+        while True:
+            tok = stream.read(1)
+            if tok == "E":
+                next = stream.read(1)
+                if next == "I":
+                    break
+                else:
+                    stream.seek(-1, 1)
+                    data += tok
+            else:
+                data += tok
+        x = readNonWhitespace(stream)
+        stream.seek(-1, 1)
+        return {"settings": settings, "data": data}
+
+    def _getData(self):
+        newdata = StringIO()
+        for operands,operator in self.operations:
+            if operator == "INLINE IMAGE":
+                newdata.write("BI")
+                dicttext = StringIO()
+                operands["settings"].writeToStream(dicttext)
+                newdata.write(dicttext.getvalue()[2:-2])
+                newdata.write("ID ")
+                newdata.write(operands["data"])
+                newdata.write("EI")
+            else:
+                for op in operands:
+                    op.writeToStream(newdata)
+                    newdata.write(" ")
+                newdata.write(operator)
+            newdata.write("\n")
+        return newdata.getvalue()
+
+    def _setData(self, value):
+        self.__parseContentStream(StringIO(value))
+
+    _data = property(_getData, _setData)
+
+
+class DocumentInformation(DictionaryObject):
+    def __init__(self):
+        DictionaryObject.__init__(self)
+
+    title = property(
+            lambda self: self.get("/Title", None),
+            None, None,
+            """The document's title, or None if not specified.  Added to pyPdf
+            in v1.6, will exist for all v1.x.""")
+
+    author = property(
+            lambda self: self.get("/Author", None),
+            None, None,
+            """The name of the person who created the document, or None if not
+            specified.  Added to pyPdf in v1.6, will exist for all v1.x.""")
+
+    subject = property(
+            lambda self: self.get("/Subject", None),
+            None, None,
+            """The subject of the document, or None if not specified.  Added to
+            pyPdf in v1.6, will exist for all v1.x.""")
+
+    creator = property(
+            lambda self: self.get("/Creator", None),
+            None, None,
+            """If the document was converted to PDF from another format, the
+            name of the application (for example, OpenOffice) that created the
+            original document from which it was converted, or None if not
+            specified.  Added to pyPdf in v1.6, will exist for all v1.x.""")
+
+    producer = property(
+            lambda self: self.get("/Producer", None),
+            None, None,
+            """If the document was converted to PDF from another format, the
+            name of the application (for example, OSX Quartz) that converted it
+            to PDF.  Added to pyPdf in v1.6, will exist for all v1.x.""")
+
 
 def convertToInt(d, size):
     if size <= 4:
@@ -661,19 +827,21 @@ def convertToInt(d, size):
 if __name__ == "__main__":
     output = PdfFileWriter()
 
-    input1 = PdfFileReader(file("..\\test\\5000-s1-05e.pdf", "rb"))
+    input1 = PdfFileReader(file("test\\5000-s1-05e.pdf", "rb"))
     page1 = input1.getPage(0)
 
-    input2 = PdfFileReader(file("..\\test\\PDFReference16.pdf", "rb"))
+    input2 = PdfFileReader(file("test\\PDFReference16.pdf", "rb"))
     page2 = input2.getPage(0)
     page3 = input2.getPage(1)
     page1.mergePage(page2)
     page1.mergePage(page3)
 
-    input3 = PdfFileReader(file("..\\test\\cc-cc.pdf", "rb"))
+    input3 = PdfFileReader(file("test\\cc-cc.pdf", "rb"))
     page1.mergePage(input3.getPage(0))
 
+    page1.compressContentStreams()
+
     output.addPage(page1)
-    output.write(file("test.pdf", "wb"))
+    output.write(file("test\\merge-test.pdf", "wb"))
 
 
